import sys
sys.path.append("{REDACTED FOR ANONYMITY}")

from datasets import Dataset
from transformers import AutoTokenizer
from tqdm import tqdm
from utils import FINETUNE_DATASET_PATH, CHAR_IIT_TASKS
import os
import csv
import json
import argparse


class CharacterTaskDataCollator:
    def __init__(self, tokenizer, input_seq_length=24, output_seq_length=16):
        self.tokenizer = tokenizer
        self.truncation = True
        self.padding = 'max_length'
        self.input_seq_length = input_seq_length
        self.output_seq_length = output_seq_length
        # 0: entailment, 1: contradiction, 2: neutral
        self.label_map = {0: '0', 1: '1', 2: '2'}

    def __call__(self, features=None):
        # Tokenize inputs and labels
        inputs_batch = self.tokenizer(features[0]['input'], padding="max_length", truncation=self.truncation,
                                      max_length=self.input_seq_length, return_tensors='pt')
        labels_batch = self.tokenizer(features[0]['output'], padding=self.padding, truncation=self.truncation,
                                      max_length=self.output_seq_length, return_tensors='pt')

        batch = {}
        batch['input_ids'] = inputs_batch["input_ids"]
        batch['labels'] = labels_batch["input_ids"]

        return batch


def save_processed_dataset(dataset, filename):
    with open(filename, 'w') as f:
        for item in dataset:
            json.dump(item, f)
            f.write('\n')


def preprocess_dataset(dataset, collator):
    for example in tqdm(dataset, total=len(dataset)):
        processed_example = collator([example])
        yield {
            'input_ids': processed_example['input_ids'].tolist(),
            'labels': processed_example['labels'].tolist(),
        }


def stream_and_preprocess(dataset, collator, task, outfile):
    # Process datasets to files
    preprocessed_dataset = preprocess_dataset(dataset, collator)

    # Create dir if it doesn't exist
    dataset_path = f"{FINETUNE_DATASET_PATH}/char_iit_json/{task}/"
    if not os.path.exists(dataset_path):
        # Create a new directory because it does not exist
        os.makedirs(dataset_path)

    save_processed_dataset(preprocessed_dataset, dataset_path + outfile)


if __name__ == '__main__':

    parser = argparse.ArgumentParser(
        description='Preprocess finetuning dataset.')

    # Toy task arguments
    parser.add_argument(
        'task',
        choices=CHAR_IIT_TASKS,
        help='Dataset to preprocess.'
    )
    args = parser.parse_args()

    # Load tokenizer
    tokenizer = AutoTokenizer.from_pretrained("google/byt5-small")

    def select_columns(filepath):
        # Initialize lists for the first two columns
        col_1 = []
        col_2 = []

        # Open the TSV file and read the data
        with open(filepath, newline='', encoding='utf-8') as tsvfile:
            reader = csv.reader(tsvfile, delimiter='\t')
            
            # Iterate over each row in the TSV file
            for row in reader:
                # Append the first two columns to the respective lists
                col_1.append(row[0])
                col_2.append(row[1])

        # Create a dictionary where each key is a column name and the value is a list of column data
        data_dict = {
            'input': col_1,
            'output': col_2
        }

        return data_dict

    CHAR_IIT_PATH = FINETUNE_DATASET_PATH + "/char_iit_data/"
    # Load the TSV file
    if args.task == 'spelling_correction':
        splits = ['train', 'val', 'test_iv', 'test_oov', 'test_real']
        collator = CharacterTaskDataCollator(tokenizer, input_seq_length=24, output_seq_length=16)
    elif args.task == 'spelling_correction_contextual':
        splits = ['train', 'val', 'test_context_dependent',
                  'test_context_independent']
        collator = CharacterTaskDataCollator(tokenizer, input_seq_length=64, output_seq_length=64)
    elif args.task == 'word_search':
        splits = ['train', 'val', 'test_oov', 'test_paraphrase', 'test_overlap', 'test_paraphrase_overlap']
        collator = CharacterTaskDataCollator(tokenizer, input_seq_length=128, output_seq_length=16)

    for split in splits:
        print("Processing spelling_correction split: ", split)
        data = select_columns(CHAR_IIT_PATH + f"{args.task}_{split}.tsv")
        dataset = Dataset.from_dict(data)
        stream_and_preprocess(
            dataset, collator, args.task, f"{args.task}_{split if split != 'val' else 'validation'}.json")
    
