from datasets import load_dataset
from transformers import AutoTokenizer
from tqdm import tqdm
from utils import FINETUNE_DATASET_PATH, FINETUNE_TASKS
import os
import json
import argparse
import sys
sys.path.append("{REDACTED FOR ANONYMITY}")


class XNLIDataCollator:
    def __init__(self, tokenizer, padding='longest', max_length=1024, truncation=True):
        self.tokenizer = tokenizer
        self.padding = padding
        self.max_length = max_length
        self.truncation = truncation
        # 0: entailment, 1: contradiction, 2: neutral
        self.label_map = {0: '0', 1: '1', 2: '2'}

    def __call__(self, features=None):
        input_text = f"P: {features[0]['premise']} H: {features[0]['hypothesis']}"
        # Convert numeric labels to text
        label_text = self.label_map[features[0]['label']]

        # Tokenize inputs and labels
        inputs_batch = self.tokenizer(input_text, padding=self.padding, truncation=self.truncation,
                                      max_length=self.max_length, return_tensors='pt')
        labels_batch = self.tokenizer(label_text, padding=self.padding, truncation=self.truncation,
                                      max_length=2, return_tensors='pt')

        batch = {}
        batch['input_ids'] = inputs_batch["input_ids"]
        batch['labels'] = labels_batch["input_ids"]

        return batch


def save_processed_dataset(dataset, filename):
    with open(filename, 'w') as f:
        for item in dataset:
            json.dump(item, f)
            f.write('\n')


def preprocess_dataset(dataset, collator):
    for example in tqdm(dataset, total=len(dataset)):
        processed_example = collator([example])
        yield {
            'input_ids': processed_example['input_ids'].tolist(),
            'labels': processed_example['labels'].tolist(),
        }


def stream_and_preprocess(dataset_name, collator, language=None, uid="", split='train'):

    # Load dataset with the split
    # Include language if specified
    if language is not None:
        dataset = load_dataset(dataset_name, language, split=split)
    else:
        dataset = load_dataset(dataset_name, split=split)

    # Process datasets to files
    preprocessed_dataset = preprocess_dataset(dataset, collator)

    # Save the processed dataset to a file
    if uid != "":
        uid = f"-{uid}"

    # Create dir if it doesn't exist
    dataset_path = f"{FINETUNE_DATASET_PATH}/{dataset_name}"
    if not os.path.exists(dataset_path):
        # Create a new directory because it does not exist
        os.makedirs(dataset_path)

    # Save the processed dataset to a file
    if language is not None:
        outfile = f"{dataset_path}/{language}-{split}{uid}.json"
    else:
        outfile = f"{dataset_path}/{split}{uid}.json"
    save_processed_dataset(preprocessed_dataset, outfile)


if __name__ == '__main__':

    parser = argparse.ArgumentParser(
        description='Preprocess finetuning dataset.')

    # Toy task arguments
    parser.add_argument(
        'dataset',
        choices=FINETUNE_TASKS,
        help='Dataset to preprocess.'
    )

    # Data arguments
    parser.add_argument('--input_seq_length', type=int,
                        default=128, help='Input sequence length for the model.')
    parser.add_argument("--uid", type=str, default="",
                        help="Unique identifier for the output files")
    parser.add_argument("--language", type=str,
                        help="Language to preprocess (if applicable)")
    args = parser.parse_args()

    # Load tokenizer
    tokenizer = AutoTokenizer.from_pretrained("google/byt5-small")

    if args.dataset == 'xnli':
        assert args.language is not None, "Language must be specified for XNLI."
        collator = XNLIDataCollator(
            tokenizer, max_length=args.input_seq_length)
    else:
        raise ValueError(
            f"Toy task must be one of {', '.join(FINETUNE_TASKS)}.")

    # Iterate over languages and sample training and validation sets
    if args.language is not None:
        print(f"Preprocessing {args.dataset} train set for {args.language}...")
        stream_and_preprocess(args.dataset, collator,
                              args.language, args.uid, "train")
        print(
            f"Preprocessing {args.dataset} validation set for {args.language}...")
        stream_and_preprocess(args.dataset, collator,
                              args.language, args.uid, "validation")
    else:
        print(f"Preprocessing {args.dataset} train set...")
        stream_and_preprocess(args.dataset, collator,
                              args.language, args.uid, "train")
        print(f"Preprocessing {args.dataset} validation set...")
        stream_and_preprocess(args.dataset, collator,
                              args.language, args.uid, "validation")
