import sys
sys.path.append("{REDACTED FOR ANONYMITY}")

import time
import os
import torch
import argparse
import pandas as pd
from tqdm import tqdm
from torch.utils.data import DataLoader
from utils import (
    get_task_dataset,
    load_model_from_path,
    CHAR_IIT_TASKS_AND_SPLITS,
)

def get_input_ids_and_labels(batch):
    input_ids = batch['input_ids'].squeeze(axis=1).to(device)
    labels = batch['labels'].squeeze(axis=1).to(device)
    return input_ids, labels

def calculate_seq_accuracy(labels, outputs):
    logits = outputs.logits
    # Get the predicted IDs
    predicted_ids = torch.argmax(logits, dim=-1)

    # Compare predicted_ids with the true labels
    correct_predictions = (predicted_ids == labels).all(dim=-1).sum().item()
    total_predictions = labels.shape[0]

    # Calculate accuracy
    accuracy = correct_predictions / total_predictions

    return accuracy


def byt5_compute_metrics(model, batch):
    # Get input ids and labels
    input_ids, labels = get_input_ids_and_labels(batch)

    # Get model outputs
    outputs = model(input_ids=input_ids,
                    labels=labels,
                    output_hidden_states=True)
    
    accuracy = calculate_seq_accuracy(labels, outputs)

    # Return accuracy and 0.0 for percent deleted tokens
    return accuracy, 0.0


def delt5_compute_metrics(model, batch):
    # Get input ids and labels
    input_ids, labels = get_input_ids_and_labels(batch)

    # Get model outputs
    outputs = model(
        input_ids=input_ids,
        labels=labels,
        hard_delete=True,
        output_hidden_states=True,
        deletion_threshold=args.deletion_threshold)

    # Get delete gate output
    delete_gate_output = outputs.delete_gate_output.squeeze(-1)

    # Calculate sequence accuracy
    accuracy = calculate_seq_accuracy(labels, outputs)

    # Compute percent deleted tokens (excluding padding tokens)
    batch_size, seq_len = input_ids.shape[0:2]
    non_pad_mask = input_ids != 0
    num_non_pad_tokens = non_pad_mask.sum()
    num_deleted_tokens = ((delete_gate_output < args.deletion_threshold) & non_pad_mask).sum()
    percent_deleted_tokens = num_deleted_tokens / num_non_pad_tokens * 100

    # Return cross entropy loss, percent deleted tokens, and new sequence length
    return accuracy, percent_deleted_tokens.item()

def load_eval_dataset(task, split, batch_size):
    # Load the dataset
    dataset = get_task_dataset(task, split, iterable_dataset=False)
    dataset = dataset.with_format(type="torch")

    # Create DataLoader for the evaluation dataset
    dataloader = DataLoader(
        dataset, batch_size=batch_size)

    return dataloader


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description='Test basic evaluation metrics for ByT5/DelT5 models.')
    parser.add_argument('model_name', type=str,
                        help='Name of model/run to load.')
    parser.add_argument('model_type',
                        type=str,
                        const='all',
                        nargs='?',
                        choices=['T5', 'DelT5', 'DecoderBaselineT5',
                                 'RandomDelT5', 'FixedDelT5'],
                        help='Type of model architecture to evaluate.')
    parser.add_argument("task", type=str, choices=CHAR_IIT_TASKS_AND_SPLITS.keys(),
                        help="Task to evaluate the model on.")
    parser.add_argument('--checkpoint', type=int,
                        default=3000, help='Model checkpoint to load for evaluation.')
    parser.add_argument('--random_seed', type=int, default=42,
                        help='Random seed for reproducibility.')
    parser.add_argument('--deletion_threshold', type=float,
                        default=-15.0, help='Deletion gate threshold.')

    # Eval arguments
    parser.add_argument('--per_device_eval_batch_size', type=int,
                        default=64, help='Batch size per device during evaluation.')


    args = parser.parse_args()

    print("Loading model...")
    model = load_model_from_path(args.model_type, model_name=args.model_name,
                                 training_task=args.task, seed=args.random_seed, ckpt=args.checkpoint)

    # Move the model to the device
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)

    # Evaluation loop
    model.eval()

    # Determine loss function based on model type
    if args.model_type == 'T5':
        compute_loss_function = byt5_compute_metrics
    elif args.model_type == 'DelT5':
        compute_loss_function = delt5_compute_metrics
    else:
        raise ValueError(
            "Model type must be 'T5' or 'DelT5'.")

    seq_accuracy_data = []
    percent_deleted_tokens_data = []
    runtime_data = []

    SPLITS = [split for split in CHAR_IIT_TASKS_AND_SPLITS[args.task] if 'test' in split]

    with torch.no_grad():
        for split in SPLITS:
            print(f"Evaluating on {args.task}, {split} split...")
            # Load the evaluation dataset
            eval_dataloader = load_eval_dataset(
                args.task, split, batch_size=args.per_device_eval_batch_size)

            # Initialize the total loss
            total_accuracy = 0.0
            total_percent_deleted_tokens = 0.0

            # Start the timer
            start_time = time.time()

            print(f"Number of batches: {len(eval_dataloader)}")
            print(f"Number of examples: {len(eval_dataloader.dataset)}")

            num_batches = len(eval_dataloader)
            for batch in tqdm(eval_dataloader):
                # Compute metrics
                acc, percent_deleted_tokens = compute_loss_function(model, batch)

                # Update the total metrics
                total_accuracy += acc
                total_percent_deleted_tokens += percent_deleted_tokens

            # End the timer
            end_time = time.time()
            eval_runtime = (end_time - start_time) / len(eval_dataloader.dataset) * 1000

            average_seq_accuracy = total_accuracy / num_batches * 100
            average_percent_deleted_tokens = total_percent_deleted_tokens / num_batches

            seq_accuracy_data.append(average_seq_accuracy)
            percent_deleted_tokens_data.append(
                average_percent_deleted_tokens)
            runtime_data.append(eval_runtime)

            print(f"Seq Accuracy: {average_seq_accuracy}")
            print(f"Percent deleted tokens: {average_percent_deleted_tokens}")
            print(f"Eval runtime: {eval_runtime} seconds")
            print()

        # Save the evaluation metrics to a CSV file
        eval_metrics = pd.DataFrame({
            'Split': SPLITS,
            'Eval Sequence Accuracy': seq_accuracy_data,
            'Eval Percent Deleted Tokens': percent_deleted_tokens_data,
            'Eval Runtime': runtime_data,
        })

        # Make directory for eval results
        dir_path = f"eval_results/char-iit-tasks/{args.task}/{args.model_type}_architecture"
        os.makedirs(dir_path, exist_ok=True)

        outfile = dir_path + f"/{args.model_name}_seed{args.random_seed}.csv"
        eval_metrics.to_csv(outfile, index=False)

        print(f"Saved evaluation metrics to: {outfile}")
